from supersuit.utils.frame_skip import check_transform_frameskip
from supersuit.utils.wrapper_chooser import WrapperChooser
from pettingzoo.utils.wrappers import BaseWrapper
import gym


class frame_skip_gym(gym.Wrapper):
    def __init__(self, env, num_frames):
        super().__init__(env)
        self.num_frames = check_transform_frameskip(num_frames)
        self.np_random, seed = gym.utils.seeding.np_random(None)

    def seed(self, seed=None):
        self.np_random, seed = gym.utils.seeding.np_random(seed)
        super().seed(seed)

    def step(self, action):
        low, high = self.num_frames
        num_skips = int(self.np_random.randint(low, high + 1))
        total_reward = 0.0

        for x in range(num_skips):
            obs, rew, done, info = super().step(action)
            total_reward += rew
            if done:
                break

        return obs, total_reward, done, info


class StepAltWrapper(BaseWrapper):
    def _modify_action(self, agent, action):
        return action

    def _modify_observation(self, agent, observation):
        return observation


class frame_skip_aec(StepAltWrapper):
    def __init__(self, env, num_frames):
        super().__init__(env)
        assert isinstance(num_frames, int), "multi-agent frame skip only takes in an integer"
        assert num_frames > 0
        check_transform_frameskip(num_frames)
        self.num_frames = num_frames

    def reset(self):
        super().reset()
        self.agents = self.env.agents[:]
        self.dones = {agent: False for agent in self.agents}
        self.rewards = {agent: 0. for agent in self.agents}
        self._cumulative_rewards = {agent: 0. for agent in self.agents}
        self.infos = {agent: {} for agent in self.agents}
        self.skip_num = {agent: 0 for agent in self.agents}
        self.old_actions = {agent: None for agent in self.agents}
        self._final_observations = {agent: None for agent in self.agents}

    def observe(self, agent):
        fin_observe = self._final_observations[agent]
        return fin_observe if fin_observe is not None else super().observe(agent)

    def step(self, action):
        self._has_updated = True
        if self.dones[self.agent_selection]:
            if self.env.agents and self.agent_selection == self.env.agent_selection:
                self.env.step(None)
            self._was_done_step(action)
            return
        cur_agent = self.agent_selection
        self._cumulative_rewards[cur_agent] = 0
        self.rewards = {a: 0. for a in self.agents}
        self.skip_num[cur_agent] = self.num_frames
        self.old_actions[cur_agent] = action
        while self.old_actions[self.env.agent_selection] is not None:
            step_agent = self.env.agent_selection
            if step_agent in self.env.dones:
                # reward = self.env.rewards[step_agent]
                # done = self.env.dones[step_agent]
                # info = self.env.infos[step_agent]
                observe, reward, done, info = self.env.last(observe=False)
                action = self.old_actions[step_agent]
                self.env.step(action)

                for agent in self.env.agents:
                    self.rewards[agent] += self.env.rewards[agent]
                self.infos[self.env.agent_selection] = info
                while self.env.agents and self.env.dones[self.env.agent_selection]:
                    done_agent = self.env.agent_selection
                    self.dones[done_agent] = True
                    self._final_observations[done_agent] = self.env.observe(done_agent)
                    self.env.step(None)
                step_agent = self.env.agent_selection

            self.skip_num[step_agent] -= 1
            if self.skip_num[step_agent] == 0:
                self.old_actions[step_agent] = None

        for agent in self.env.agents:
            self.dones[agent] = self.env.dones[agent]
            self.infos[agent] = self.env.infos[agent]
        self.agent_selection = self.env.agent_selection
        self._accumulate_rewards()
        self._dones_step_first()


frame_skip_v0 = WrapperChooser(aec_wrapper=frame_skip_aec, gym_wrapper=frame_skip_gym)



import gym
from pettingzoo.utils.wrappers import OrderEnforcingWrapper as PettingzooWrap
from supersuit.utils.wrapper_chooser import WrapperChooser
from supersuit.utils.base_parallel_wrapper import ParallelWraper


class shared_wrapper_aec(PettingzooWrap):
    def __init__(self, env, modifier_class):
        super().__init__(env)

        self.modifiers = {}
        self.observation_spaces = {}
        self.action_spaces = {}
        for agent in self.env.possible_agents:
            self.modifiers[agent] = modifier_class()
            self.observation_spaces[agent] = self.modifiers[agent].modify_obs_space(self.env.observation_spaces[agent])
            self.action_spaces[agent] = self.modifiers[agent].modify_action_space(self.env.action_spaces[agent])

    def seed(self, seed=None):
        super().seed(seed)
        for agent, mod in sorted(self.modifiers.items()):
            mod.seed(seed)
            if seed is not None:
                seed += 1

    def reset(self):
        for mod in self.modifiers.values():
            mod.reset()
        super().reset()
        self.modifiers[self.agent_selection].modify_obs(super().observe(self.agent_selection))

    def step(self, action):
        mod = self.modifiers[self.agent_selection]
        action = mod.modify_action(action)
        if self.dones[self.agent_selection]:
            action = None
        super().step(action)
        self.modifiers[self.agent_selection].modify_obs(super().observe(self.agent_selection))

    def observe(self, agent):
        return self.modifiers[agent].get_last_obs()


class shared_wrapper_parr(ParallelWraper):
    def __init__(self, env, modifier_class):
        super().__init__(env)

        self.modifiers = {}
        self.observation_spaces = {}
        self.action_spaces = {}
        for agent in self.env.possible_agents:
            self.modifiers[agent] = modifier_class()
            self.observation_spaces[agent] = self.modifiers[agent].modify_obs_space(self.env.observation_spaces[agent])
            self.action_spaces[agent] = self.modifiers[agent].modify_action_space(self.env.action_spaces[agent])

    def seed(self, seed=None):
        super().seed(seed)
        for agent, mod in sorted(self.modifiers.items()):
            mod.seed(seed)
            if seed is not None:
                seed += 1

    def reset(self):
        observations = super().reset()
        for agent, mod in self.modifiers.items():
            mod.reset()
            observations[agent] = mod.modify_obs(observations[agent])
        return observations

    def step(self, actions):
        actions = {agent: self.modifiers[agent].modify_action(action) for agent, action in actions.items()}
        observations, rewards, dones, infos = super().step(actions)
        observations = {agent: self.modifiers[agent].modify_obs(obs) for agent, obs in observations.items()}
        return observations, rewards, dones, infos


class shared_wrapper_gym(gym.Wrapper):
    def __init__(self, env, modifier_class):
        super().__init__(env)
        self.modifier = modifier_class()
        self.observation_space = self.modifier.modify_obs_space(self.observation_space)
        self.action_space = self.modifier.modify_action_space(self.action_space)

    def seed(self, seed=None):
        super().seed(seed)
        self.modifier.seed(seed)

    def reset(self):
        self.modifier.reset()
        obs = super().reset()
        obs = self.modifier.modify_obs(obs)
        return obs

    def step(self, action):
        obs, rew, done, info = super().step(self.modifier.modify_action(action))
        obs = self.modifier.modify_obs(obs)
        return obs, rew, done, info


shared_wrapper = WrapperChooser(aec_wrapper=shared_wrapper_aec, gym_wrapper=shared_wrapper_gym, parallel_wrapper=shared_wrapper_parr)



from .frame_skip import frame_skip_v0 # NOQA
from .basic_wrappers import color_reduction_v0, resize_v0, dtype_v0, \
    flatten_v0, reshape_v0, normalize_obs_v0, clip_actions_v0, clip_reward_v0, \
    scale_actions_v0 # NOQA
from .nan_wrappers import nan_random_v0, nan_noop_v0, nan_zeros_v0 # NOQA

from .delay_observations import delay_observations_v0 # NOQA
from .frame_stack import frame_stack_v1, frame_stack_v0 # NOQA
from .max_observation import max_observation_v0 # NOQA
from .sticky_actions import sticky_actions_v0 # NOQA



from .utils.shared_array import SharedArray
from .utils.space_wrapper import SpaceWrapper
import multiprocessing as mp
import numpy as np
import traceback
import gym.vector


def compress_info(infos):
    non_empty_infs = [(i, info) for i, info in enumerate(infos) if info]
    return non_empty_infs


def decompress_info(num_envs, idx_starts, comp_infos):
    all_info = [{}] * num_envs
    for idx_start, comp_infos in zip(idx_starts, comp_infos):
        for i, info in comp_infos:
            all_info[idx_start + i] = info
    return all_info


def async_loop(vec_env_constr, inpt_p, pipe, shared_obs, shared_actions, shared_rews, shared_dones):
    inpt_p.close()
    try:
        vec_env = vec_env_constr()

        pipe.send((vec_env.num_envs))
        env_start_idx = pipe.recv()
        env_end_idx = env_start_idx + vec_env.num_envs
        while True:
            instr = pipe.recv()
            comp_infos = []
            if instr == "reset":
                obs = vec_env.reset()
                shared_obs.np_arr[env_start_idx:env_end_idx] = obs
                shared_dones.np_arr[env_start_idx:env_end_idx] = False
                shared_rews.np_arr[env_start_idx:env_end_idx] = 0.0
            elif instr == "step":
                actions = shared_actions.np_arr[env_start_idx:env_end_idx]
                observations, rewards, dones, infos = vec_env.step(actions)
                shared_obs.np_arr[env_start_idx:env_end_idx] = observations
                shared_dones.np_arr[env_start_idx:env_end_idx] = dones
                shared_rews.np_arr[env_start_idx:env_end_idx] = rewards
                comp_infos = compress_info(infos)
            elif instr == "close":
                vec_env.close()
            elif isinstance(instr, tuple):
                name, data = instr
                if name == "seed":
                    vec_env.seed(data)
                elif name == "env_is_wrapped":
                    comp_infos = vec_env.env_is_wrapped(data)
                elif name == "render":
                    render_result = vec_env.render(data)
                    if data == "rgb_array":
                        comp_infos = render_result
                else:
                    raise AssertionError("bad tuple instruction name: " + name)
            elif instr == "terminate":
                return
            else:
                raise AssertionError("bad instruction: " + instr)
            pipe.send(comp_infos)
    except BaseException as e:
        tb = traceback.format_exc()
        pipe.send((e, tb))


class ProcConcatVec(gym.vector.VectorEnv):
    def __init__(self, vec_env_constrs, observation_space, action_space, tot_num_envs, metadata):
        self.observation_space = observation_space
        self.action_space = action_space
        self.num_envs = num_envs = tot_num_envs
        self.metadata = metadata

        self.shared_obs = SharedArray((num_envs,) + self.observation_space.shape, dtype=self.observation_space.dtype)
        act_space_wrap = SpaceWrapper(self.action_space)
        self.shared_act = SharedArray((num_envs,) + act_space_wrap.shape, dtype=act_space_wrap.dtype)
        self.shared_rews = SharedArray((num_envs,), dtype=np.float32)
        self.shared_dones = SharedArray((num_envs,), dtype=np.uint8)

        pipes = []
        procs = []
        for constr in vec_env_constrs:
            inpt, outpt = mp.Pipe()
            constr = gym.vector.async_vector_env.CloudpickleWrapper(constr)
            proc = mp.Process(
                target=async_loop, args=(constr, inpt, outpt, self.shared_obs, self.shared_act, self.shared_rews, self.shared_dones)
            )
            proc.start()
            outpt.close()
            pipes.append(inpt)
            procs.append(proc)

        self.pipes = pipes
        self.procs = procs

        num_envs = 0
        env_nums = self._receive_info()
        idx_starts = []
        for pipe, cnum_env in zip(self.pipes, env_nums):
            cur_env_idx = num_envs
            num_envs += cnum_env
            pipe.send(cur_env_idx)
            idx_starts.append(cur_env_idx)

        assert num_envs == tot_num_envs
        self.idx_starts = idx_starts

    def reset(self):
        for pipe in self.pipes:
            pipe.send("reset")

        self._receive_info()

        observations = self.shared_obs.np_arr
        return observations

    def step_async(self, actions):
        self.shared_act.np_arr[:] = actions
        for pipe in self.pipes:
            pipe.send("step")

    def _receive_info(self):
        all_data = []
        for cin in self.pipes:
            data = cin.recv()
            if isinstance(data, tuple):
                e, tb = data
                print(tb)
                raise e
            all_data.append(data)
        return all_data

    def step_wait(self):
        compressed_infos = self._receive_info()
        infos = decompress_info(self.num_envs, self.idx_starts, compressed_infos)
        observations = self.shared_obs.np_arr
        rewards = self.shared_rews.np_arr
        dones = self.shared_dones.np_arr
        return observations, rewards, dones, infos

    def step(self, actions):
        self.step_async(actions)
        return self.step_wait()

    def seed(self, seed=None):
        for i, pipe in enumerate(self.pipes):
            pipe.send(("seed", seed + self.idx_starts[i]))

        self._receive_info()

    def __del__(self):
        for pipe in self.pipes:
            try:
                pipe.send("terminate")
            except ConnectionError:
                pass
        for proc in self.procs:
            proc.join()

    def render(self, mode="human"):
        self.pipes[0].send(("render", mode))
        render_result = self.pipes[0].recv()

        if isinstance(render_result, tuple):
            e, tb = render_result
            print(tb)
            raise e

        if mode == "rgb_array":
            return render_result

    def close(self):
        for pipe in self.pipes:
            pipe.send("close")
        for pipe in self.pipes:
            try:
                pipe.recv()
            except EOFError:
                raise RuntimeError("only one multiproccessing vector environment can open a window over the duration of a process")
            except ConnectionError:
                pass

    def env_is_wrapped(self, wrapper_class, indices=None):
        for i, pipe in enumerate(self.pipes):
            pipe.send(("env_is_wrapped", wrapper_class))

        results = self._receive_info()
        return sum(results, [])



from supersuit.utils.base_aec_wrapper import BaseWrapper
from gym.spaces import Box
import numpy as np
import gym
from supersuit.utils.wrapper_chooser import WrapperChooser
from supersuit.utils.deprecated import Deprecated


class ObservationWrapper(BaseWrapper):
    def _modify_action(self, agent, action):
        return action


class black_death_aec(ObservationWrapper):
    def _check_wrapper_params(self):
        for space in self.observation_spaces.values():
            assert isinstance(space, gym.spaces.Box), f"observation sapces for black death must be Box spaces, is {space}"

    def _modify_spaces(self):
        self.observation_spaces = {agent: Box(low=np.minimum(0, obs.low), high=np.maximum(0, obs.high), dtype=obs.dtype) for agent, obs in self.observation_spaces.items()}

    def observe(self, agent):
        return np.zeros_like(self.observation_spaces[agent].low) if agent not in self.env.dones else self.env.observe(agent)

    def reset(self):
        super().reset()
        self._agent_idx = 0
        self.agent_selection = self.possible_agents[self._agent_idx]
        self.agents = self.possible_agents[:]
        self._cumulative_rewards = {agent: 0 for agent in self.agents}
        self._update_items()

    def _update_items(self):
        self.dones = {}
        self.infos = {}
        self.rewards = {}

        _env_finishing = self._agent_idx == len(self.possible_agents) - 1 and all(self.env.dones.values())

        for agent in self.agents:
            self.dones[agent] = _env_finishing
            self.rewards[agent] = self.env.rewards.get(agent, 0)
            self.infos[agent] = self.env.infos.get(agent, {})
            self._cumulative_rewards[agent] += self.env.rewards.get(agent, 0)

            # self._cumulative_rewards[agent] = self.env._cumulative_rewards.get(agent, 0)

    def step(self, action):
        self._has_updated = True
        if self.dones[self.agent_selection]:
            return self._was_done_step(action)

        cur_agent = self.agent_selection
        self._cumulative_rewards[cur_agent] = 0
        if cur_agent == self.env.agent_selection:
            assert cur_agent in self.env.dones
            if self.env.dones[cur_agent]:
                action = None

            self.env.step(action)

        self._update_items()

        self._agent_idx = (1 + self._agent_idx) % len(self.possible_agents)
        self.agent_selection = self.possible_agents[self._agent_idx]

        self._dones_step_first()


black_death_v0 = Deprecated("black_death", "v0", "v2")
black_death_v1 = Deprecated("black_death", "v1", "v2")
black_death_v2 = WrapperChooser(aec_wrapper=black_death_aec)



from stable_baselines3.common.vec_env import VecEnvWrapper


class SB3VecEnvWrapper(VecEnvWrapper):
    def __init__(self, venv):
        self.venv = venv
        self.num_envs = venv.num_envs
        self.observation_space = venv.observation_space
        self.action_space = venv.action_space

    def reset(self):
        return self.venv.reset()

    def step_wait(self):
        return self.venv.step_wait()

    def env_is_wrapped(self, wrapper_class, indices=None):
        # ignores indicies
        return self.venv.env_is_wrapped(wrapper_class)

    def getattr_recursive(self, name):
        raise AttributeError(name)

    def getattr_depth_check(self, name, already_found):
        return None



from pettingzoo.utils.env import ParallelEnv


class ParallelWraper(ParallelEnv):
    def __init__(self, env):
        self.env = env
        self.observation_spaces = env.observation_spaces
        self.action_spaces = env.action_spaces
        self.metadata = env.metadata
        self.possible_agents = env.possible_agents

    def reset(self):
        res = self.env.reset()
        self.agents = self.env.agents
        return res

    def step(self, actions):
        res = self.env.step(actions)
        self.agents = self.env.agents
        return res

    def render(self, mode="human"):
        return self.env.render(mode)

    def close(self):
        return self.env.close()



from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper
from gym.spaces import Box, Discrete
from supersuit.utils.frame_stack import stack_obs_space, stack_init, stack_obs
from supersuit.utils.deprecated import Deprecated


def frame_stack_v1(env, stack_size=4):
    assert isinstance(stack_size, int), "stack size of frame_stack must be an int"

    class FrameStackModifier(BaseModifier):
        def modify_obs_space(self, obs_space):
            if isinstance(obs_space, Box):
                assert 1 <= len(obs_space.shape) <= 3, "frame_stack only works for 1, 2 or 3 dimensional observations"
            elif isinstance(obs_space, Discrete):
                pass
            else:
                assert False, "Stacking is currently only allowed for Box and Discrete observation spaces. The given observation space is {}".format(obs_space)

            self.old_obs_space = obs_space
            self.observation_space = stack_obs_space(obs_space, stack_size)
            return self.observation_space

        def reset(self):
            self.stack = stack_init(self.old_obs_space, stack_size)

        def modify_obs(self, obs):
            self.stack = stack_obs(
                self.stack,
                obs,
                self.old_obs_space,
                stack_size,
            )
            return self.stack

        def get_last_obs(self):
            return self.stack

    return shared_wrapper(env, FrameStackModifier)


frame_stack_v0 = Deprecated("frame_stack", "v0", "v1")



from gym.spaces import Box, Discrete
from .dummy_aec_env import DummyEnv
import numpy as np
from supersuit import (
    frame_stack_v1,
    reshape_v0,
    observation_lambda_v0,
    action_lambda_v1,
    pad_action_space_v0,
    pad_observations_v0,
    dtype_v0,
    scale_actions_v0,
)
import supersuit
import pytest
from pettingzoo.utils.wrappers import OrderEnforcingWrapper as PettingzooWrap


base_obs = {"a{}".format(idx): np.zeros([8, 8, 3], dtype=np.float32) + np.arange(3) + idx for idx in range(2)}
base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[8, 8, 3]) for idx in range(2)}
base_act_spaces = {"a{}".format(idx): Discrete(5) for idx in range(2)}


def test_frame_stack():
    base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[2, 3]) for idx in range(2)}
    base_obs = {"a{}".format(idx): np.zeros([2, 3]) + np.arange(3) + idx for idx in range(2)}
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = frame_stack_v1(base_env, 4)
    obs = env.reset()
    obs, _, _, _ = env.last()
    assert obs.shape == (2, 3, 4)
    env.step(2)
    first_obs, _, _, _ = env.last()
    print(first_obs[:, :, -1])
    assert np.all(np.equal(first_obs[:, :, -1], base_obs["a1"]))
    assert np.all(np.equal(first_obs[:, :, :-1], 0))

    base_obs = {"a{}".format(idx): idx + 3 for idx in range(2)}
    base_env = DummyEnv(base_obs, base_act_spaces, base_act_spaces)
    env = frame_stack_v1(base_env, 4)
    obs = env.reset()
    obs, _, _, _ = env.last()
    assert env.observation_spaces[env.agent_selection].n == 5 ** 4
    env.step(2)
    first_obs, _, _, _ = env.last()
    assert first_obs == 4
    env.step(2)
    second_obs, _, _, _ = env.last()
    assert second_obs == 3 + 3 * 5
    for x in range(8):
        nth_obs = env.step(2)
        nth_obs, _, _, _ = env.last()
    assert nth_obs == ((3 * 5 + 3) * 5 + 3) * 5 + 3


def test_frame_skip():
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = supersuit.frame_skip_v0(base_env, 3)
    env.reset()
    for a in env.agent_iter(8):
        obs, rew, done, info = env.last()
        env.step(0 if not done else None)


def test_agent_indicator():
    let = ["a", "a", "b"]
    base_obs = {"{}_{}".format(let[idx], idx): np.zeros([2, 3]) for idx in range(3)}
    base_obs_space = {"{}_{}".format(let[idx], idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[2, 3]) for idx in range(3)}
    base_act_spaces = {"{}_{}".format(let[idx], idx): Discrete(5) for idx in range(3)}

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = supersuit.agent_indicator_v0(base_env, type_only=True)
    env.reset()
    obs, _, _, _ = env.last()
    assert obs.shape == (2, 3, 3)
    assert env.observation_spaces["a_0"].shape == (2, 3, 3)
    first_obs = env.step(2)

    env = supersuit.agent_indicator_v0(base_env, type_only=False)
    env.reset()
    obs, _, _, _ = env.last()
    assert obs.shape == (2, 3, 4)
    assert env.observation_spaces["a_0"].shape == (2, 3, 4)
    env.step(2)


def test_reshape():
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = reshape_v0(base_env, (64, 3))
    env.reset()
    obs, _, _, _ = env.last()
    assert obs.shape == (64, 3)
    env.step(5)
    first_obs, _, _, _ = env.last()
    assert np.all(np.equal(first_obs, base_obs["a1"].reshape([64, 3])))


def new_continuous_dummy():

    base_obs = {"a_{}".format(idx): (np.zeros([8, 8, 3], dtype=np.float32) + np.arange(3) + idx).astype(np.float32) for idx in range(2)}
    base_obs_space = {"a_{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[8, 8, 3]) for idx in range(2)}
    base_act_spaces = {"a_{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[3]) for idx in range(2)}

    return PettingzooWrap(DummyEnv(base_obs, base_obs_space, base_act_spaces))


def new_dummy():

    base_obs = {"a_{}".format(idx): (np.zeros([8, 8, 3], dtype=np.float32) + np.arange(3) + idx).astype(np.float32) for idx in range(2)}
    base_obs_space = {"a_{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[8, 8, 3]) for idx in range(2)}
    base_act_spaces = {"a_{}".format(idx): Discrete(5) for idx in range(2)}

    return PettingzooWrap(DummyEnv(base_obs, base_obs_space, base_act_spaces))


wrappers = [
    supersuit.color_reduction_v0(new_dummy(), "R"),
    supersuit.resize_v0(dtype_v0(new_dummy(), np.uint8), x_size=5, y_size=10),
    supersuit.resize_v0(dtype_v0(new_dummy(), np.uint8), x_size=5, y_size=10, linear_interp=True),
    supersuit.dtype_v0(new_dummy(), np.int32),
    supersuit.flatten_v0(new_dummy()),
    supersuit.reshape_v0(new_dummy(), (64, 3)),
    supersuit.normalize_obs_v0(new_dummy(), env_min=-1, env_max=5.0),
    supersuit.frame_stack_v1(new_dummy(), 8),
    supersuit.pad_observations_v0(new_dummy()),
    supersuit.pad_action_space_v0(new_dummy()),
    supersuit.black_death_v2(new_dummy()),
    supersuit.agent_indicator_v0(new_dummy(), True),
    supersuit.agent_indicator_v0(new_dummy(), False),
    supersuit.reward_lambda_v0(new_dummy(), lambda x: x / 10),
    supersuit.clip_reward_v0(new_dummy()),
    supersuit.clip_actions_v0(new_continuous_dummy()),
    supersuit.scale_actions_v0(new_continuous_dummy(), 0.3),
    supersuit.frame_skip_v0(new_dummy(), 4),
    supersuit.sticky_actions_v0(new_dummy(), 0.75),
    supersuit.delay_observations_v0(new_dummy(), 3),
    supersuit.max_observation_v0(new_dummy(), 3),
    supersuit.nan_noop_v0(new_dummy(), 0),
    supersuit.nan_zeros_v0(new_dummy()),
    supersuit.nan_random_v0(new_dummy()),
]


@pytest.mark.parametrize("env", wrappers)
def test_basic_wrappers(env):
    env.seed(5)
    env.reset()
    obs, _, _, _ = env.last()
    act_space = env.action_spaces[env.agent_selection]
    obs_space = env.observation_spaces[env.agent_selection]
    first_obs = env.observe("a_0")
    assert obs_space.contains(first_obs)
    assert first_obs.dtype == obs_space.dtype
    env.step(act_space.sample())
    for agent in env.agent_iter():
        act_space = env.action_spaces[env.agent_selection]
        env.step(act_space.sample() if not env.dones[agent] else None)


def test_rew_lambda():
    env = supersuit.reward_lambda_v0(new_dummy(), lambda x: x / 10)
    env.reset()
    assert env.rewards[env.agent_selection] == 1.0 / 10


def test_observation_lambda():
    def add1(obs, obs_space):
        return obs + 1

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = observation_lambda_v0(base_env, add1)
    env.reset()
    obs0, _, _, _ = env.last()
    assert int(obs0[0][0][0]) == 1
    env = observation_lambda_v0(env, add1)
    env.reset()
    obs0, _, _, _ = env.last()
    assert int(obs0[0][0][0]) == 2

    def tile_obs(obs, obs_space):
        shape_size = len(obs.shape)
        tile_shape = [1] * shape_size
        tile_shape[0] *= 2
        return np.tile(obs, tile_shape)

    env = observation_lambda_v0(env, tile_obs)
    env.reset()
    obs0, _, _, _ = env.last()
    assert env.observation_spaces[env.agent_selection].shape == (16, 8, 3)

    def change_shape_fn(obs_space):
        return Box(low=0, high=1, shape=(32, 8, 3))

    env = observation_lambda_v0(env, tile_obs)
    env.reset()
    obs0, _, _, _ = env.last()
    assert env.observation_spaces[env.agent_selection].shape == (32, 8, 3)
    assert obs0.shape == (32, 8, 3)

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = observation_lambda_v0(base_env, lambda obs, obs_space, agent: obs + base_env.possible_agents.index(agent))
    env.reset()
    obs0 = env.observe(env.agents[0])
    obs1 = env.observe(env.agents[1])

    assert int(obs0[0][0][0]) == 0
    assert int(obs1[0][0][0]) == 2
    assert (env.observation_spaces[env.agents[0]].high + 1 == env.observation_spaces[env.agents[1]].high).all()

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = observation_lambda_v0(base_env,
                                lambda obs, obs_space, agent: obs + base_env.possible_agents.index(agent),
                                lambda obs_space, agent: Box(obs_space.low, obs_space.high + base_env.possible_agents.index(agent)))
    env.reset()
    obs0 = env.observe(env.agents[0])
    obs1 = env.observe(env.agents[1])

    assert int(obs0[0][0][0]) == 0
    assert int(obs1[0][0][0]) == 2
    assert (env.observation_spaces[env.agents[0]].high + 1 == env.observation_spaces[env.agents[1]].high).all()


def test_action_lambda():
    def inc1(x, space):
        return x + 1

    def change_space_fn(space):
        return Discrete(space.n + 1)

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = action_lambda_v1(base_env, inc1, change_space_fn)
    env.reset()
    env.step(5)

    def one_hot(x, n):
        v = np.zeros(n)
        v[x] = 1
        return v

    act_spaces = {"a{}".format(idx): Box(low=0, high=1, shape=(15,)) for idx in range(2)}
    base_env = DummyEnv(base_obs, base_obs_space, act_spaces)
    env = action_lambda_v1(
        base_env,
        lambda action, act_space: one_hot(action, act_space.shape[0]),
        lambda act_space: Discrete(act_space.shape[0]),
    )
    assert env.action_spaces[env.possible_agents[0]].n == 15

    env.reset()
    env.step(2)

    act_spaces = {"a{}".format(idx): Box(low=0, high=1, shape=(15,)) for idx in range(2)}
    base_env = DummyEnv(base_obs, base_obs_space, act_spaces)
    env = action_lambda_v1(
        base_env,
        lambda action, act_space, agent: one_hot(action + base_env.possible_agents.index(agent), act_space.shape[0]),
        lambda act_space, agent: Discrete(act_space.shape[0] + base_env.possible_agents.index(agent)),
    )
    assert env.action_spaces[env.possible_agents[0]].n == 15
    assert env.action_spaces[env.possible_agents[1]].n == 16
    env.reset()
    env.step(2)


def test_dehomogenize():
    base_act_spaces = {"a{}".format(idx): Discrete(5 + idx) for idx in range(2)}

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = pad_action_space_v0(base_env)
    env.reset()
    assert all([s.n == 6 for s in env.action_spaces.values()])
    env.step(5)


class DummyNaNEnv(DummyEnv):
    def step(self, action):
        super().step(action)
        assert (not (np.isnan(action))), "Action was a NaN, even though it should not have been."


def test_NaN_noop():
    base_env = DummyNaNEnv(base_obs, base_obs_space, base_act_spaces)
    wrapped_env = supersuit.nan_noop_v0(base_env, 0)
    wrapped_env.reset()
    wrapped_env.step(np.nan)


def test_NaN_zeros():
    base_env = DummyNaNEnv(base_obs, base_obs_space, base_act_spaces)
    wrapped_env = supersuit.nan_zeros_v0(base_env)
    wrapped_env.reset()
    wrapped_env.step(np.nan)


def test_NaN_random():
    base_env = DummyNaNEnv(base_obs, base_obs_space, base_act_spaces)
    wrapped_env = supersuit.nan_random_v0(base_env)
    wrapped_env.reset()
    wrapped_env.step(np.nan)


class DummyScaleEnv(DummyEnv):
    def step(self, action):
        self._observations[self.agent_selection] = action
        super().step(action)


def test_scale_action_wrapper():
    base_obs = {"a{}".format(idx): np.zeros([8, 8, 3], dtype=np.float32) + np.arange(3) + idx for idx in range(2)}
    base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[8, 8, 3]) for idx in range(2)}
    base_act_spaces = {"a{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[3]) for idx in range(2)}
    base_env = DummyScaleEnv(base_obs, base_obs_space, base_act_spaces)
    wrapped_env = scale_actions_v0(base_env, 2)
    wrapped_env.reset()
    wrapped_env.step(np.array([2, 1, 3], dtype=np.float32))
    scaled_action = wrapped_env.observe(wrapped_env.agents[0])
    assert (scaled_action == np.array([4, 2, 6], dtype=np.float32)).all()

    with pytest.raises(AssertionError):
        base_act_spaces = {"a{}".format(idx): Discrete(5) for idx in range(2)}
        base_env = DummyScaleEnv(base_obs, base_obs_space, base_act_spaces)
        wrapped_env = scale_actions_v0(base_env, 2)



from .agent_indication import agent_indicator_v0 # NOQA
from .black_death import black_death_v2, black_death_v1, black_death_v0 # NOQA
from .padding_wrappers import pad_action_space_v0, pad_observations_v0 # NOQA



from gym.spaces import Space
from supersuit.utils.base_aec_wrapper import BaseWrapper
from supersuit.utils.wrapper_chooser import WrapperChooser
from supersuit.utils.deprecated import Deprecated
import gym


class aec_action_lambda(BaseWrapper):
    def __init__(self, env, change_action_fn, change_space_fn):
        assert callable(change_action_fn), "change_action_fn needs to be a function. It is {}".format(change_action_fn)
        assert callable(change_space_fn), "change_space_fn needs to be a function. It is {}".format(change_space_fn)

        old_space_fn = change_space_fn
        old_action_fn = change_action_fn

        def space_fn_ignore(space, agent):
            return old_space_fn(space)

        def action_fn_ignore(action, space, agent):
            return old_action_fn(action, space)

        agent0 = env.possible_agents[0]
        agent0_space = env.action_spaces[agent0]

        try:
            new_space0 = change_space_fn(agent0_space, agent0)
        except TypeError:
            change_space_fn = space_fn_ignore
            new_space0 = change_space_fn(agent0_space, agent0)

        try:
            change_action_fn(new_space0.sample(), agent0_space, agent0)
        except TypeError:
            change_action_fn = action_fn_ignore

        self.change_action_fn = change_action_fn
        self.change_space_fn = change_space_fn

        super().__init__(env)

    def _modify_observation(self, agent, observation):
        return observation

    def _modify_spaces(self):
        new_spaces = {}
        for agent, space in self.action_spaces.items():
            new_spaces[agent] = self.change_space_fn(space, agent)
            assert isinstance(new_spaces[agent], Space), "output of change_space_fn argument to action_lambda_wrapper must be a gym space"

        self.action_spaces = new_spaces

    def _modify_action(self, agent, action):
        return self.change_action_fn(action, self.env.action_spaces[agent], agent)


class gym_action_lambda(gym.Wrapper):
    def __init__(self, env, change_action_fn, change_space_fn):
        assert callable(change_action_fn), "change_action_fn needs to be a function. It is {}".format(change_action_fn)
        assert callable(change_space_fn), "change_space_fn needs to be a function. It is {}".format(change_space_fn)
        self.change_action_fn = change_action_fn
        self.change_space_fn = change_space_fn

        super().__init__(env)
        self._modify_spaces()

    def _modify_spaces(self):
        new_space = self.change_space_fn(self.action_space)
        assert isinstance(new_space, Space), "output of change_space_fn argument to action_lambda_wrapper must be a gym space"
        self.action_space = new_space

    def _modify_action(self, action):
        return self.change_action_fn(action, self.env.action_space)

    def step(self, action):
        return super().step(self._modify_action(action))


action_lambda_v0 = Deprecated("action_lambda", "v0", "v1")
action_lambda_v1 = WrapperChooser(aec_wrapper=aec_action_lambda, gym_wrapper=gym_action_lambda)



from .single_vec_env import SingleVecEnv # NOQA
from .multiproc_vec import ProcConcatVec # NOQA
from .concat_vec_env import ConcatVecEnv # NOQA
from .markov_vector_wrapper import MarkovVectorEnv # NOQA
from .constructors import MakeCPUAsyncConstructor # NOQA



import gym
import numpy as np
from gym.spaces import Box, Space
from supersuit.utils.base_aec_wrapper import BaseWrapper
from supersuit.utils.wrapper_chooser import WrapperChooser


class aec_observation_lambda(BaseWrapper):
    def __init__(self, env, change_observation_fn, change_obs_space_fn=None):
        assert callable(change_observation_fn), "change_observation_fn needs to be a function. It is {}".format(change_observation_fn)
        assert change_obs_space_fn is None or callable(change_obs_space_fn), "change_obs_space_fn needs to be a function. It is {}".format(change_obs_space_fn)

        old_space_fn = change_obs_space_fn
        old_obs_fn = change_observation_fn

        def space_fn_ignore(space, agent):
            return old_space_fn(space)

        def obs_fn_ignore(obs, obs_space, agent):
            return old_obs_fn(obs, obs_space)

        agent0 = env.possible_agents[0]
        agent0_space = env.observation_spaces[agent0]

        if change_obs_space_fn is not None:
            try:
                change_obs_space_fn(agent0_space, agent0)
            except TypeError:
                change_obs_space_fn = space_fn_ignore

        try:
            change_observation_fn(agent0_space.sample(), agent0_space, agent0)
        except TypeError:
            change_observation_fn = obs_fn_ignore

        self.change_observation_fn = change_observation_fn
        self.change_obs_space_fn = change_obs_space_fn

        super().__init__(env)

    def _modify_action(self, agent, action):
        return action

    def _check_wrapper_params(self):
        if self.change_obs_space_fn is None:
            spaces = self.observation_spaces.values()
            for space in spaces:
                assert isinstance(space, Box), "the observation_lambda_wrapper only allows the change_obs_space_fn argument to be optional for Box observation spaces"

    def _modify_spaces(self):
        new_spaces = {}
        for agent, space in self.observation_spaces.items():
            if self.change_obs_space_fn is None:
                trans_low = self.change_observation_fn(space.low, space, agent)
                trans_high = self.change_observation_fn(space.high, space, agent)
                new_low = np.minimum(trans_low, trans_high)
                new_high = np.maximum(trans_low, trans_high)

                new_spaces[agent] = Box(low=new_low, high=new_high, dtype=new_low.dtype)
            else:
                new_space = self.change_obs_space_fn(space, agent)
                assert isinstance(new_space, Space), "output of change_obs_space_fn to observation_lambda_wrapper must be a gym space"
                new_spaces[agent] = new_space
        self.observation_spaces = new_spaces

    def _modify_observation(self, agent, observation):
        return self.change_observation_fn(observation, self.env.observation_spaces[agent], agent)


class gym_observation_lambda(gym.Wrapper):
    def __init__(self, env, change_observation_fn, change_obs_space_fn=None):
        assert callable(change_observation_fn), "change_observation_fn needs to be a function. It is {}".format(change_observation_fn)
        assert change_obs_space_fn is None or callable(change_obs_space_fn), "change_obs_space_fn needs to be a function. It is {}".format(change_obs_space_fn)
        self.change_observation_fn = change_observation_fn
        self.change_obs_space_fn = change_obs_space_fn

        super().__init__(env)
        self._check_wrapper_params()
        self._modify_spaces()

    def _check_wrapper_params(self):
        if self.change_obs_space_fn is None:
            space = self.observation_space
            assert isinstance(space, Box), "the observation_lambda_wrapper only allows the change_obs_space_fn argument to be optional for Box observation spaces"

    def _modify_spaces(self):
        space = self.observation_space

        if self.change_obs_space_fn is None:
            new_low = self.change_observation_fn(space.low, space)
            new_high = self.change_observation_fn(space.high, space)
            new_space = Box(low=new_low, high=new_high, dtype=new_low.dtype)
        else:
            new_space = self.change_obs_space_fn(space)
            assert isinstance(new_space, Space), "output of change_obs_space_fn to observation_lambda_wrapper must be a gym space"
        self.observation_space = new_space

    def _modify_observation(self, observation):
        return self.change_observation_fn(observation, self.env.observation_space)

    def step(self, action):
        observation, rew, done, info = self.env.step(action)
        observation = self._modify_observation(observation)
        return observation, rew, done, info

    def reset(self):
        observation = self.env.reset()
        observation = self._modify_observation(observation)
        return observation


observation_lambda_v0 = WrapperChooser(aec_wrapper=aec_observation_lambda, gym_wrapper=gym_observation_lambda)



from supersuit import gym_vec_env_v0
import gym
import numpy as np


def test_vec_env_args():
    env = gym.make("Acrobot-v1")
    num_envs = 8
    vec_env = gym_vec_env_v0(env, num_envs)
    vec_env.reset()
    obs, rew, dones, infos = vec_env.step([0] + [1] * (vec_env.num_envs - 1))
    assert not np.any(np.equal(obs[0], obs[1]))


def test_all_vec_env_fns():
    num_envs = 8
    env = gym.make("Acrobot-v1")
    vec_env = gym_vec_env_v0(env, num_envs, False)
    vec_env = gym_vec_env_v0(env, num_envs, True)



import gym
import copy


class DummyEnv(gym.Env):
    def __init__(self, observation, observation_space, action_space):
        super().__init__()
        self._observation = observation
        self.observation_space = observation_space
        self.action_space = action_space

    def step(self, action):
        return self._observation, 1, False, {}

    def reset(self):
        return self._observation



import copy
import multiprocessing as mp
from pettingzoo.utils.agent_selector import agent_selector
import numpy as np
import ctypes
import gym
from .base_aec_vec_env import VectorAECEnv
import warnings
import signal
import traceback


class SpaceWrapper:
    def __init__(self, space):
        if isinstance(space, gym.spaces.Discrete):
            self.shape = ()
            self.dtype = np.dtype(np.int64)
            self.low = 0
        elif isinstance(space, gym.spaces.Box):
            self.shape = space.shape
            self.dtype = np.dtype(space.dtype)
            self.low = space.low
        else:
            assert False, "ProcVectorEnv only support Box and Discrete types"


class SharedData:
    def __init__(self, array, num_envs, shape, dtype):
        self.array = array
        self.num_envs = num_envs
        self.np_arr = np.frombuffer(array, dtype=dtype).reshape((num_envs,) + shape)


def create_shared_data(num_envs, obs_space, act_space):
    obs_array = mp.Array(np.ctypeslib.as_ctypes_type(obs_space.dtype), int(num_envs * np.prod(obs_space.shape)), lock=False)
    act_array = mp.Array(np.ctypeslib.as_ctypes_type(act_space.dtype), int(num_envs * np.prod(act_space.shape)), lock=False)
    rew_array = mp.Array(ctypes.c_float, num_envs, lock=False)
    cum_rew_array = mp.Array(ctypes.c_float, num_envs, lock=False)
    done_array = mp.Array(ctypes.c_bool, num_envs, lock=False)
    data = (obs_array, act_array, rew_array, cum_rew_array, done_array)
    return data


def create_env_data(num_envs):
    env_dones = mp.Array(ctypes.c_bool, num_envs, lock=False)
    agent_sel_idx = mp.Array(ctypes.c_uint, num_envs, lock=False)
    return env_dones, agent_sel_idx


class AgentSharedData:
    def __init__(self, num_envs, obs_space, act_space, data):
        self.num_envs = num_envs

        obs_array, act_array, rew_array, cum_rew_array, done_array = data
        self.obs = SharedData(obs_array, num_envs, obs_space.shape, obs_space.dtype)
        self.act = SharedData(act_array, num_envs, act_space.shape, act_space.dtype)
        self.rewards = SharedData(rew_array, num_envs, (), np.float32)
        self._cumulative_rewards = SharedData(cum_rew_array, num_envs, (), np.float32)
        self.dones = SharedData(done_array, num_envs, (), np.uint8)


class EnvSharedData:
    def __init__(self, num_envs, data):
        env_dones, agent_idx_array = data
        self.env_dones = SharedData(env_dones, num_envs, (), np.uint8)
        self.agent_sel_idx = SharedData(agent_idx_array, num_envs, (), np.uint32)


class _SeperableAECWrapper:
    def __init__(self, env_constructors, num_envs):
        self.envs = [env_constructor() for env_constructor in env_constructors]
        self.env = self.envs[0]
        self.possible_agents = self.env.possible_agents
        self.agent_indexes = {agent: i for i, agent in enumerate(self.env.possible_agents)}
        self.dead_obss = {agent: np.zeros_like(SpaceWrapper(obs_space).low) for agent, obs_space in self.env.observation_spaces.items()}

    def reset(self):
        for env in self.envs:
            env.reset()

        self.rewards = {agent: [env.rewards.get(agent, 0) for env in self.envs] for agent in self.possible_agents}
        self._cumulative_rewards = {agent: [env._cumulative_rewards.get(agent, 0) for env in self.envs] for agent in self.possible_agents}
        self.dones = {agent: [env.dones.get(agent, True) for env in self.envs] for agent in self.possible_agents}
        self.infos = {agent: [env.infos.get(agent, {}) for env in self.envs] for agent in self.possible_agents}

    def observe(self, agent):
        observations = []
        for env in self.envs:
            observations.append(env.observe(agent) if agent in env.dones else self.dead_obss[agent])
        return observations

    def seed(self, seed=None):
        for i, env in enumerate(self.envs):
            env.seed(seed + i)

    def step(self, agent_step, actions):
        assert len(actions) == len(self.envs)

        env_dones = []
        for act, env in zip(actions, self.envs):
            env_done = not env.agents
            env_dones.append(env_done)
            if env_done:
                env.reset()
            elif env.agent_selection == agent_step:
                if env.dones[agent_step]:
                    act = None
                env.step(act)

        self.rewards = {agent: [env.rewards.get(agent, 0) for env in self.envs] for agent in self.possible_agents}
        self._cumulative_rewards = {agent: [env._cumulative_rewards.get(agent, 0) for env in self.envs] for agent in self.possible_agents}
        self.dones = {agent: [env.dones.get(agent, True) for env in self.envs] for agent in self.possible_agents}
        self.infos = {agent: [env.infos.get(agent, {}) for env in self.envs] for agent in self.possible_agents}

        return env_dones

    def get_agent_indexes(self):
        return [self.agent_indexes[env.agent_selection] for env in self.envs]


def sig_handle(signal_object, argvar):
    raise RuntimeError()


def init_parallel_env():
    signal.signal(signal.SIGINT, sig_handle)
    signal.signal(signal.SIGTERM, sig_handle)


def write_out_data(rewards, cumulative_rews, dones, num_envs, start_index, shared_data):
    for agent in shared_data:
        rews = np.asarray(rewards[agent], dtype=np.float32)
        cum_rews = np.asarray(cumulative_rews[agent], dtype=np.float32)
        dns = np.asarray(dones[agent], dtype=np.uint8)
        cur_data = shared_data[agent]
        cur_data.rewards.np_arr[start_index : start_index + num_envs] = rews
        cur_data._cumulative_rewards.np_arr[start_index : start_index + num_envs] = cum_rews
        cur_data.dones.np_arr[start_index : start_index + num_envs] = dns


def write_env_data(env_dones, indexes, num_envs, start_index, shared_data):
    shared_data.env_dones.np_arr[start_index : start_index + num_envs] = env_dones
    agent_indexes = np.asarray(indexes, dtype=np.uint32)
    shared_data.agent_sel_idx.np_arr[start_index : start_index + num_envs] = agent_indexes


def write_obs(obs, num_env, start_index, shared_data):
    for i, o in enumerate(obs):
        if o is not None:
            shared_data.obs.np_arr[start_index + i] = o


def compress_info(infos):
    all_infos = {}
    for agent, infs in infos.items():
        non_empty_infs = [(i, info) for i, info in enumerate(infs) if info]
        if non_empty_infs:
            all_infos[agent] = non_empty_infs
    return all_infos


def decompress_info(agents, num_envs, idx_starts, comp_infos):
    all_info = {agent: [{}] * num_envs for agent in agents}
    for idx_start, comp_info in zip(idx_starts, comp_infos):
        for agent, inf_data in comp_info.items():
            agent_info = all_info[agent]
            for i, info in inf_data:
                agent_info[idx_start + i] = info
    return all_info


def env_worker(env_constructors, total_num_envs, idx_start, my_num_envs, agent_arrays, env_arrays, pipe):
    try:
        env = _SeperableAECWrapper(env_constructors, my_num_envs)
        shared_datas = {
            agent: AgentSharedData(
                total_num_envs, SpaceWrapper(env.env.observation_spaces[agent]), SpaceWrapper(env.env.action_spaces[agent]), agent_arrays[agent]
            )
            for agent in env.possible_agents
        }

        env_datas = EnvSharedData(total_num_envs, env_arrays)

        while True:
            instruction, data = pipe.recv()
            if instruction == "reset":
                env.reset()
                write_out_data(env.rewards, env._cumulative_rewards, env.dones, my_num_envs, idx_start, shared_datas)
                env_dones = np.zeros(my_num_envs, dtype=np.uint8)
                write_env_data(env_dones, env.get_agent_indexes(), my_num_envs, idx_start, env_datas)

                pipe.send(compress_info(env.infos))

            elif instruction == "observe":
                agent_observe = data
                obs = env.observe(agent_observe)
                write_obs(obs, my_num_envs, idx_start, shared_datas[agent_observe])

                pipe.send(True)
            elif instruction == "step":
                step_agent, do_observe = data

                actions = shared_datas[step_agent].act.np_arr[idx_start : idx_start + my_num_envs]

                env_dones = env.step(step_agent, actions)
                write_out_data(env.rewards, env._cumulative_rewards, env.dones, my_num_envs, idx_start, shared_datas)
                write_env_data(env_dones, env.get_agent_indexes(), my_num_envs, idx_start, env_datas)

                pipe.send(compress_info(env.infos))
            elif instruction == "seed":
                env.seed(data)
                pipe.send(True)
            elif instruction == "terminate":
                return
            else:
                assert False, "Bad instruction sent to ProcVectorEnv worker"
    except Exception as e:
        tb = traceback.format_exc()
        pipe.send((e, tb))


class AsyncAECVectorEnv(VectorAECEnv):
    def __init__(self, env_constructors, num_cpus=None, return_copy=True):
        # set signaling so that crashing is handled gracefully
        init_parallel_env()

        num_envs = len(env_constructors)

        if num_cpus is None:
            num_cpus = mp.cpu_count()

        num_cpus = min(num_cpus, num_envs)
        assert num_envs > 0

        assert num_envs >= 1
        assert callable(env_constructors[0]), "env_constructor must be a callable object (i.e function) that create an environment"
        # self.envs = [env_constructor() for _ in range(num_envs)]
        self.env = env = env_constructors[0]()
        self.max_num_agents = len(self.env.possible_agents)
        self.possible_agents = self.env.possible_agents
        self.observation_spaces = copy.copy(self.env.observation_spaces)
        self.action_spaces = copy.copy(self.env.action_spaces)
        self.order_is_nondeterministic = False
        self.num_envs = num_envs

        self.agent_indexes = {agent: i for i, agent in enumerate(self.env.possible_agents)}

        self._agent_selector = agent_selector(self.possible_agents)

        all_arrays = {
            agent: create_shared_data(
                num_envs,
                SpaceWrapper(self.observation_spaces[agent]),
                SpaceWrapper(self.action_spaces[agent]),
            )
            for agent in self.possible_agents
        }

        self.shared_datas = {
            agent: AgentSharedData(num_envs, SpaceWrapper(env.observation_spaces[agent]), SpaceWrapper(env.action_spaces[agent]), all_arrays[agent])
            for agent in env.possible_agents
        }

        env_arrays = create_env_data(num_envs)

        self.env_datas = EnvSharedData(num_envs, env_arrays)
        self.return_copy = return_copy

        self.procs = []
        self.pipes = [mp.Pipe() for _ in range(num_cpus)]
        self.con_ins = [con_in for con_in, con_out in self.pipes]
        self.con_outs = [con_out for con_in, con_out in self.pipes]
        self.env_starts = []
        env_counter = 0
        for pidx in range(num_cpus):
            envs_left = num_envs - env_counter
            allocated_envs = min(envs_left, (num_envs + num_cpus - 1) // num_cpus)
            proc_constructors = env_constructors[env_counter : env_counter + allocated_envs]
            proc = mp.Process(
                target=env_worker, args=(proc_constructors, num_envs, env_counter, allocated_envs, all_arrays, env_arrays, self.con_outs[pidx])
            )
            self.procs.append(proc)
            self.env_starts.append(env_counter)

            proc.start()
            env_counter += allocated_envs

    def _find_active_agent(self):
        cur_selection = self.agent_selection
        while not np.any(np.equal(self.agent_indexes[cur_selection], self.env_datas.agent_sel_idx.np_arr)):
            cur_selection = self._agent_selector.next()
        return cur_selection

    def _receive_info(self):
        all_data = []
        for cin in self.con_ins:
            data = cin.recv()
            if isinstance(data, tuple):
                err, tb = data
                print(tb)
                raise RuntimeError("ProcVectorEnv received error from subprocess (look up there ^^^ for error)")
            all_data.append(data)
        return all_data

    def copy(self, data):
        if self.return_copy:
            return np.copy(data)
        else:
            return data

    def _load_next_data(self, reset):
        all_compressed_info = self._receive_info()

        all_info = decompress_info(self.possible_agents, self.num_envs, self.env_starts, all_compressed_info)

        self.agent_selection = self._agent_selector.reset() if reset else self._agent_selector.next()
        self.agent_selection = self._find_active_agent()

        passes = np.not_equal(self.env_datas.agent_sel_idx.np_arr, self.agent_indexes[self.agent_selection])

        assert not np.all(passes), "something went wrong with finding agent"
        if np.any(passes) or self.order_is_nondeterministic:
            warnings.warn("The agent order of sub-environments of ProcVectorEnv differs, likely due to agent death. The ProcVectorEnv only returns one agent at a time, so it will now 'pass' environments where the current agent is not active, taking up to O(n) more time")
            self.order_is_nondeterministic = True

        self.dones = {agent: self.copy(self.shared_datas[agent].dones.np_arr) for agent in self.possible_agents}
        self.rewards = {agent: self.copy(self.shared_datas[agent].rewards.np_arr) for agent in self.possible_agents}
        self._cumulative_rewards = {agent: self.copy(self.shared_datas[agent]._cumulative_rewards.np_arr) for agent in self.possible_agents}
        self.infos = all_info
        env_dones = self.env_datas.env_dones.np_arr
        self.passes = self.copy(passes)
        self.env_dones = self.copy(env_dones)

    def last(self, observe=True):
        last_agent = self.agent_selection
        obs = self.observe(last_agent) if observe else None
        return obs, self._cumulative_rewards[last_agent], self.dones[last_agent], self.env_dones, self.passes, self.infos[last_agent]

    def reset(self, observe=True):
        for cin in self.con_ins:
            cin.send(("reset", observe))

        self._load_next_data(True)

    def step(self, actions, observe=True):
        step_agent = self.agent_selection

        self.shared_datas[self.agent_selection].act.np_arr[:] = actions
        for cin in self.con_ins:
            cin.send(("step", (step_agent, observe)))

        self._load_next_data(False)

    def observe(self, agent):
        for cin in self.con_ins:
            cin.send(("observe", agent))

        # wait until all are finished
        self._receive_info()

        obs = self.copy(self.shared_datas[self.agent_selection].obs.np_arr)
        return obs

    def seed(self, seed):
        for start, cin in zip(self.env_starts, self.con_ins):
            cin.send(("seed", seed + start if seed is not None else None))

        self._receive_info()

    def __del__(self):
        for cin in self.con_ins:
            cin.send(("terminate", None))
        for proc in self.procs:
            proc.join()




class BaseModifier:
    def __init__(self):
        pass

    def reset(self):
        pass

    def modify_obs(self, obs):
        self.cur_obs = obs
        return obs

    def seed(self, seed):
        pass

    def get_last_obs(self):
        return self.cur_obs

    def modify_obs_space(self, obs_space):
        self.observation_space = obs_space
        return obs_space

    def modify_action(self, act):
        return act

    def modify_action_space(self, act_space):
        self.action_space = act_space
        return act_space



import copy
from pettingzoo.utils.agent_selector import agent_selector
import numpy as np
from .base_aec_vec_env import VectorAECEnv


class SyncAECVectorEnv(VectorAECEnv):
    def __init__(self, env_constructors):
        assert len(env_constructors) >= 1
        assert callable(env_constructors[0]), "env_constructor must be a callable object (i.e function) that create an environment"

        self.envs = [env_constructor() for env_constructor in env_constructors]
        self.num_envs = len(env_constructors)
        self.env = self.envs[0]
        self.max_num_agents = self.env.max_num_agents
        self.possible_agents = self.env.possible_agents
        self.observation_spaces = copy.copy(self.env.observation_spaces)
        self.action_spaces = copy.copy(self.env.action_spaces)
        self._agent_selector = agent_selector(self.possible_agents)

    def _find_active_agent(self):
        cur_selection = self.agent_selection
        while not any(cur_selection == env.agent_selection for env in self.envs):
            cur_selection = self._agent_selector.next()
        return cur_selection

    def _collect_dicts(self):
        self.rewards = {
            agent: np.array([env.rewards[agent] if agent in env.rewards else 0 for env in self.envs], dtype=np.float32)
            for agent in self.possible_agents
        }
        self._cumulative_rewards = {
            agent: np.array([env._cumulative_rewards[agent] if agent in env._cumulative_rewards else 0 for env in self.envs], dtype=np.float32)
            for agent in self.possible_agents
        }
        self.dones = {
            agent: np.array([env.dones[agent] if agent in env.dones else True for env in self.envs], dtype=np.uint8) for agent in self.possible_agents
        }
        self.infos = {agent: [env.infos[agent] if agent in env.infos else {} for env in self.envs] for agent in self.possible_agents}

    def reset(self):
        """
        returns: list of observations
        """
        for env in self.envs:
            env.reset()

        self.agent_selection = self._agent_selector.reset()
        self.agent_selection = self._find_active_agent()

        self._collect_dicts()
        self.envs_dones = np.zeros(self.num_envs)

    def seed(self, seed=None):
        for i, env in enumerate(self.envs):
            env.seed(seed + i)

    def observe(self, agent):
        observations = []
        for env in self.envs:
            obs = env.observe(agent) if agent in env.dones else np.zeros_like(self.observation_spaces[agent].low)
            observations.append(obs)
        return np.stack(observations)

    def last(self, observe=True):
        passes = np.array([env.agent_selection != self.agent_selection for env in self.envs], dtype=np.uint8)
        last_agent = self.agent_selection
        obs = self.observe(last_agent) if observe else None
        return obs, self._cumulative_rewards[last_agent], self.dones[last_agent], self.envs_dones, passes, self.infos[last_agent]

    def step(self, actions, observe=True):
        assert len(actions) == len(self.envs)
        old_agent = self.agent_selection

        envs_dones = []
        for i, (act, env) in enumerate(zip(actions, self.envs)):
            env_done = not env.agents
            envs_dones.append(env_done)
            if env_done:
                env.reset()
            elif env.agent_selection == old_agent:
                env.step(act if not self.dones[old_agent][i] else None)

        self.agent_selection = self._agent_selector.next()
        self.agent_selection = self._find_active_agent()

        self.envs_dones = np.array(envs_dones)
        self._collect_dicts()



import gym
import cloudpickle
from . import MakeCPUAsyncConstructor, MarkovVectorEnv
from pettingzoo.utils.env import ParallelEnv
import warnings


def vec_env_args(env, num_envs):
    def env_fn():
        env_copy = cloudpickle.loads(cloudpickle.dumps(env))
        env.seed(None)
        return env_copy

    return [env_fn] * num_envs, env.observation_space, env.action_space


def warn_not_gym_env(env, fn_name):
    if not isinstance(env, gym.Env):
        warnings.warn(f"{fn_name} took in an environment which does not inherit from gym.Env. Note that gym_vec_env only takes in gym-style environments, not pettingzoo environments.")


def gym_vec_env_v0(env, num_envs, multiprocessing=False):
    warn_not_gym_env(env, "gym_vec_env")
    args = vec_env_args(env, num_envs)
    constructor = gym.vector.AsyncVectorEnv if multiprocessing else gym.vector.SyncVectorEnv
    return constructor(*args)


def stable_baselines_vec_env_v0(env, num_envs, multiprocessing=False):
    import stable_baselines

    warn_not_gym_env(env, "stable_baselines_vec_env")
    args = vec_env_args(env, num_envs)[:1]
    constructor = stable_baselines.common.vec_env.SubprocVecEnv if multiprocessing else stable_baselines.common.vec_env.DummyVecEnv
    return constructor(*args)


def stable_baselines3_vec_env_v0(env, num_envs, multiprocessing=False):
    import stable_baselines3

    warn_not_gym_env(env, "stable_baselines3_vec_env")
    args = vec_env_args(env, num_envs)[:1]
    constructor = stable_baselines3.common.vec_env.SubprocVecEnv if multiprocessing else stable_baselines3.common.vec_env.DummyVecEnv
    return constructor(*args)


def concat_vec_envs_v0(vec_env, num_vec_envs, num_cpus=0, base_class='gym'):
    num_cpus = min(num_cpus, num_vec_envs)
    vec_env = MakeCPUAsyncConstructor(num_cpus)(*vec_env_args(vec_env, num_vec_envs))

    if base_class == "gym":
        return vec_env
    elif base_class == "stable_baselines":
        from .sb_vector_wrapper import SBVecEnvWrapper
        return SBVecEnvWrapper(vec_env)
    elif base_class == "stable_baselines3":
        from .sb3_vector_wrapper import SB3VecEnvWrapper
        return SB3VecEnvWrapper(vec_env)
    else:
        raise ValueError("supersuit_vec_env only supports 'gym', 'stable_baselines', and 'stable_baselines3' for its base_class")


def pettingzoo_env_to_vec_env_v0(parallel_env):
    assert isinstance(parallel_env, ParallelEnv), "pettingzoo_env_to_vec_env takes in a pettingzoo ParallelEnv. Can create a parallel_env with pistonball.parallel_env() or convert it from an AEC env with `from pettingzoo.utils.conversions import to_parallel; to_parallel(env)``"
    return MarkovVectorEnv(parallel_env)



from supersuit.utils import basic_transforms
from supersuit.lambda_wrappers import action_lambda_v1, observation_lambda_v0, reward_lambda_v0
import numpy as np
from gym.spaces import Box


def basic_obs_wrapper(env, module, param):
    def change_space(space):
        module.check_param(space, param)
        space = module.change_obs_space(space, param)
        return space

    def change_obs(obs, obs_space):
        return module.change_observation(obs, obs_space, param)
    return observation_lambda_v0(env, change_obs, change_space)


def color_reduction_v0(env, mode="full"):
    return basic_obs_wrapper(env, basic_transforms.color_reduction, mode)


def resize_v0(env, x_size, y_size, linear_interp=False):
    scale_tuple = (x_size, y_size, linear_interp)
    return basic_obs_wrapper(env, basic_transforms.resize, scale_tuple)


def dtype_v0(env, dtype):
    return basic_obs_wrapper(env, basic_transforms.dtype, dtype)


def flatten_v0(env):
    return basic_obs_wrapper(env, basic_transforms.flatten, True)


def reshape_v0(env, shape):
    return basic_obs_wrapper(env, basic_transforms.reshape, shape)


def normalize_obs_v0(env, env_min=0.0, env_max=1.0):
    return basic_obs_wrapper(env, basic_transforms.normalize_obs, (env_min, env_max))


def clip_actions_v0(env):
    return action_lambda_v1(env,
        lambda action, act_space: np.clip(action, act_space.low, act_space.high),
        lambda act_space: act_space)


def scale_actions_v0(env, scale):
    def change_act_space(act_space):
        assert isinstance(act_space, Box), "scale_actions_v0 only works with a Box action space"
        return Box(low=act_space.low * scale, high=act_space.high * scale)
    return action_lambda_v1(env,
        lambda action, act_space: np.asarray(action) * scale,
        lambda act_space: change_act_space(act_space))


def clip_reward_v0(env, lower_bound=-1, upper_bound=1):
    return reward_lambda_v0(env, lambda rew: max(min(rew, upper_bound), lower_bound))
